package main

import (
	"bufio"
	"encoding/json"
	"flag"
	"fmt"
	"os"
	"strings"

	tea "github.com/charmbracelet/bubbletea"
	charmlog "github.com/charmbracelet/log"
	"gitlab.com/tanna.dev/renovate-pretty-log/internal/log"
	"gitlab.com/tanna.dev/renovate-pretty-log/internal/style"
)

// parentViewportHeight indicates how many lines of logs will be shown on a given "page" of logs
const parentViewportHeight = 30

// extrasViewportHeight indicates how many lines of "extra" metadata will be shown under a given log, if expanded
const extrasViewportHeight = 20

type model struct {
	// logs contains the log lines themselves. They must be pre-filtered
	logs   []log.Line
	cursor int
	// parentViewportStart is the index of the first visible log
	parentViewportStart int
	// expanded tracks which indexes are currently expanded
	expanded map[int]bool
	// extrasExpandedViewport allows expanding a given log.Line's Extras, to show any additional context the log line has
	extrasExpandedViewport extrasExpandedViewport
}

func newModel(lines []log.Line) model {
	return model{
		logs:                lines,
		cursor:              0,
		parentViewportStart: 0,
		expanded:            make(map[int]bool),
	}
}

func (m model) Init() tea.Cmd {
	return nil
}

func (m model) Update(msg tea.Msg) (tea.Model, tea.Cmd) {
	switch msg := msg.(type) {
	case tea.KeyMsg:
		switch msg.String() {
		case
			"up",
			"k":
			m.MoveUp()
		case
			"down",
			"j":
			m.MoveDown()
		case
			"ctrl+u",
			"pgup":
			m.cursor -= parentViewportHeight
			if m.cursor < 0 {
				m.cursor = 0
			}
			m.parentViewportStart -= parentViewportHeight
			if m.parentViewportStart < 0 {
				m.parentViewportStart = 0
			}

		case
			"ctrl+d",
			"pgdown":
			m.cursor += parentViewportHeight

			if m.cursor >= len(m.logs)-1 {
				m.cursor = len(m.logs) - 1
			}

			m.parentViewportStart += parentViewportHeight
			if m.parentViewportStart >= len(m.logs)-1 {
				m.parentViewportStart = len(m.logs) - 1
			}
		case "enter":
			m.expanded[m.cursor] = !m.expanded[m.cursor]

			log := m.logs[m.cursor]
			extras := strings.Split(style.RenderLineExtras(log), "\n")
			m.extrasExpandedViewport.cursor = 0
			m.extrasExpandedViewport.length = len(extras)
		case "q", "esc", "ctrl+c":
			return m, tea.Quit
		}
	}

	return m, nil
}

func (m model) View() string {
	var b strings.Builder

	end := min(m.parentViewportStart+parentViewportHeight, len(m.logs))

	for i := m.parentViewportStart; i < end; i++ {
		line := m.logs[i]

		cursor := " "
		if i == m.cursor {
			cursor = ">"
		}

		hasExtras := len(line.Extras) > 0
		hasExtrasIndicator := " "
		if hasExtras {
			hasExtrasIndicator = style.LogLineHasExtrasStyle.Render("+")
			if m.expanded[i] {
				hasExtrasIndicator = style.LogLineHasExtrasStyle.Render("-")
			}
		}

		renderedLine := style.RenderLine(line, false)

		fmt.Fprintf(&b, "%s %s %s", cursor, hasExtrasIndicator, renderedLine)
		if m.expanded[i] && hasExtras {
			extras := strings.Split(style.RenderLineExtras(line), "\n")

			tooLongPrefix := ""
			tooLongSuffix := ""

			start := 0
			end := extrasViewportHeight
			if end > len(extras) {
				end = len(extras)
			}

			if len(extras) > extrasViewportHeight {
				tooLongSuffix = "\n" + style.LogLineExtraIndent + "..."
			}

			if i == m.cursor {
				if m.extrasExpandedViewport.AtTop() {
					tooLongPrefix = ""
				} else {
					tooLongPrefix = fmt.Sprintf("%s... %d/%d\n", style.LogLineExtraIndent, m.extrasExpandedViewport.cursor, m.extrasExpandedViewport.length)
				}

				if m.extrasExpandedViewport.AtBottom() {
					tooLongSuffix = ""
				} else {
					tooLongSuffix = fmt.Sprintf("\n%s... %d/%d", style.LogLineExtraIndent, m.extrasExpandedViewport.cursor+extrasViewportHeight, m.extrasExpandedViewport.length)
				}

				start = m.extrasExpandedViewport.cursor
				end = m.extrasExpandedViewport.cursor + extrasViewportHeight
				if end > len(extras) {
					end = len(extras)
				}
			}

			fmt.Fprintf(&b, "\n%s%s%s",
				tooLongPrefix,
				style.LogLineExtraValueStyle.Render(strings.Join(extras[start:end], "\n")),
				tooLongSuffix)
		}
		fmt.Fprintln(&b)
	}

	if len(m.logs) > end {
		fmt.Fprintf(&b, "\nViewing %d/%d logs\n", m.cursor, len(m.logs)-end)
	}

	// TODO: migrate this to a help widget
	currentLog := m.logs[m.cursor]
	if len(currentLog.Extras) > 0 {
		if m.expanded[m.cursor] {
			fmt.Fprintf(&b, "\n[Use arrow keys to navigate, %s, q to quit]", style.LogLineHasExtrasStyle.Render("Enter to collapse"))
		} else {
			fmt.Fprintf(&b, "\n[Use arrow keys to navigate, %s, q to quit]", style.LogLineHasExtrasStyle.Render("Enter to expand"))
		}
	} else {
		fmt.Fprintf(&b, "\n[Use arrow keys to navigate, %s, q to quit]", style.LogLineTimestampStyle.Render("Enter to expand"))
	}

	return b.String()
}

func (m *model) MoveUp() {
	// if we're currently at an expanded log line ...
	if m.expanded[m.cursor] {
		// and we're already at the top
		if m.extrasExpandedViewport.AtTop() {
			// then scroll up as normal, out of the expanded log line
			m.cursor--
			if m.cursor < 0 {
				m.cursor = 0
			}

			// if we've gone out of view (as we're now before the previous start point) scroll us up
			if m.cursor < m.parentViewportStart {
				m.parentViewportStart--
			}
			if m.parentViewportStart < 0 {
				m.parentViewportStart = 0
			}

			// we may now be on a log line that is expanded, so we need to makes sure our internal state is reset
			m.updateExtrasExpandedViewportStateAfterMoving()
		} else {
			// otherwise, we want to scroll our expanded log line up a line
			m.extrasExpandedViewport.MoveUp()
		}
	} else if m.cursor > 0 {
		// if we're on a regular line, then move us up
		m.cursor--
		if m.cursor < 0 {
			m.cursor = 0
		}

		// if we've gone out of view (as we're now before the previous start point) scroll us up
		if m.cursor < m.parentViewportStart {
			m.parentViewportStart--
		}
		if m.parentViewportStart < 0 {
			m.parentViewportStart = 0
		}

		previous := m.cursor - 1
		if previous < 0 {
			previous = 0
		}

		// we may now be on a log line that is expanded, so we need to makes sure our internal state is reset
		m.updateExtrasExpandedViewportStateAfterMoving()
	}
}

func (m *model) MoveDown() {
	// if we're currently at an expanded log line ...
	if m.expanded[m.cursor] {
		// and we're already at the bottom
		if m.extrasExpandedViewport.AtBottom() {
			// then scroll down as normal, out of the expanded log line
			m.cursor++
			if m.cursor > len(m.logs)-1 {
				m.cursor = len(m.logs) - 1
			}

			// if we've gone out of view (as we're now before the previous start point) scroll us down
			if m.cursor > m.parentViewportStart {
				m.parentViewportStart++
			}
			if m.parentViewportStart > len(m.logs)-1 {
				m.parentViewportStart = len(m.logs) - 1
			}

			// we may now be on a log line that is expanded, so we need to makes sure our internal state is reset
			m.updateExtrasExpandedViewportStateAfterMoving()
		} else {
			// otherwise, we want to scroll our expanded log line down a line
			m.extrasExpandedViewport.MoveDown()
		}
	} else if m.cursor < len(m.logs)-1 {
		// if we're on a regular line, then move us down
		m.cursor++
		if m.cursor > len(m.logs)-1 {
			m.cursor = len(m.logs) - 1
		}

		// if we've gone out of view (as we're now before the previous start point) scroll us down
		if m.cursor > m.parentViewportStart {
			m.parentViewportStart++
		}
		if m.parentViewportStart > len(m.logs)-1 {
			m.parentViewportStart = len(m.logs) - 1
		}

		// we may now be on a log line that is expanded, so we need to makes sure our internal state is reset
		m.updateExtrasExpandedViewportStateAfterMoving()
	}
}

func (m *model) updateExtrasExpandedViewportStateAfterMoving() {
	log := m.logs[m.cursor]
	extras := strings.Split(style.RenderLineExtras(log), "\n")
	m.extrasExpandedViewport.cursor = 0
	m.extrasExpandedViewport.length = len(extras)
}

// extrasExpandedViewport allows expanding a given log line's "extra" log metadata, in a scrollable fashion.
type extrasExpandedViewport struct {
	cursor int
	length int
}

func (e extrasExpandedViewport) AtTop() bool {
	return e.cursor == 0
}

func (e extrasExpandedViewport) AtBottom() bool {
	if e.cursor == (e.length - 1) {
		return true
	}

	isShowingLastPage := (e.length)-extrasViewportHeight <= e.cursor

	return isShowingLastPage
}

func (e *extrasExpandedViewport) MoveUp() {
	e.cursor--
	if e.cursor <= 0 {
		e.cursor = 0
	}
}

func (e *extrasExpandedViewport) MoveDown() {
	e.cursor++
	if e.cursor > e.length {
		e.cursor = e.length
	}
}

func parseLevelFlag(level *string) (log.Level, error) {
	if level == nil {
		return log.LevelInfo, nil
	}

	val := strings.ToLower(*level)

	switch val {
	case "trace":
		return log.LevelTrace, nil
	case "debug":
		return log.LevelDebug, nil
	case "info":
		return log.LevelInfo, nil
	case "warn":
		return log.LevelWarn, nil
	case "error":
		return log.LevelError, nil
	case "fatal":
		return log.LevelFatal, nil
	}

	return -1, fmt.Errorf("unknown log level: %v", *level)
}

func main() {
	cliLogger := charmlog.New(os.Stderr)

	levelFlag := flag.String("level", "info", "Set the log level (trace, debug, info, warn, error, fatal)")
	pathFlag := flag.String("path", "", "The path to the debug log file (`RENOVATE_DEBUG_LOG_FILE`)")
	flag.Parse()

	if pathFlag == nil {
		cliLogger.Fatal("No `-path` flag was provided")
		os.Exit(1)
	}

	if *pathFlag == "" {
		cliLogger.Fatal("No `-path` flag was provided")
		os.Exit(1)
	}

	f, err := os.Open(*pathFlag)
	if err != nil {
		cliLogger.Fatal(fmt.Sprintf("Error opening file: %v", err), "err", err)
		os.Exit(1)
	}

	defer f.Close()

	allowedLogLevel, err := parseLevelFlag(levelFlag)
	if err != nil {
		cliLogger.Fatal(fmt.Sprintf("The `-level` flag was invalid: %v", err), "err", err)
		os.Exit(1)
	}

	var lines []log.Line
	// TODO via https://gitlab.com/jamietanna/dotfiles-arch/-/blob/67d79966/go/home/go/src/jvt.me/dotfiles/jjj/main.go#L24-26 we may want to allow handling /very/ long log lines
	scanner := bufio.NewScanner(f)
	for scanner.Scan() {
		var line log.Line
		data := scanner.Bytes()
		err := json.Unmarshal(data, &line)
		if err != nil {
			cliLogger.Fatal(fmt.Sprintf("Failed to parse a log line as JSON: %v", err), "err", err)
			continue
		}

		if line.Level >= allowedLogLevel {
			lines = append(lines, line)
		}
	}

	if len(lines) == 0 {
		fmt.Printf("The file you provided did not contain any %s level logs\n", allowedLogLevel.String())
		os.Exit(0)
	}

	p := tea.NewProgram(newModel(lines))
	if _, err := p.Run(); err != nil {
		fmt.Fprintf(os.Stderr, "Error starting app: %v\n", err)
		os.Exit(1)
	}
}
