package main

import (
	"bufio"
	"encoding/json"
	"flag"
	"fmt"
	"os"
	"strings"

	charmlog "github.com/charmbracelet/log"

	log "gitlab.com/tanna.dev/renovate-pretty-log/internal/log"
	"gitlab.com/tanna.dev/renovate-pretty-log/internal/style"
)

func parseLevelFlag(level *string) (log.Level, error) {
	if level == nil {
		return log.LevelInfo, nil
	}

	val := strings.ToLower(*level)

	switch val {
	case "trace":
		return log.LevelTrace, nil
	case "debug":
		return log.LevelDebug, nil
	case "info":
		return log.LevelInfo, nil
	case "warn":
		return log.LevelWarn, nil
	case "error":
		return log.LevelError, nil
	case "fatal":
		return log.LevelFatal, nil
	}

	return -1, fmt.Errorf("unknown log level: %v", *level)
}

func main() {
	cliLogger := charmlog.New(os.Stderr)

	levelFlag := flag.String("level", "info", "Set the log level (trace, debug, info, warn, error, fatal)")
	pathFlag := flag.String("path", "", "The path to the debug log file (`RENOVATE_DEBUG_LOG_FILE`)")
	flag.Parse()

	if pathFlag == nil {
		cliLogger.Fatal("No `-path` flag was provided")
		os.Exit(1)
	}

	f, err := os.Open(*pathFlag)
	if err != nil {
		cliLogger.Fatal(fmt.Sprintf("Error opening file: %v", err), "err", err)
		os.Exit(1)
	}

	allowedLogLevel, err := parseLevelFlag(levelFlag)
	if err != nil {
		cliLogger.Fatal(fmt.Sprintf("The `-level` flag was invalid: %v", err), "err", err)
		os.Exit(1)
	}

	defer f.Close()
	scanner := bufio.NewScanner(f)
	for scanner.Scan() {
		var line log.Line
		data := scanner.Bytes()
		err := json.Unmarshal(data, &line)
		if err != nil {
			cliLogger.Fatal(fmt.Sprintf("Failed to parse a log line as JSON: %v", err), "err", err)
			continue
		}

		if line.Level < allowedLogLevel {
			cliLogger.Debug("Skipping log line from a log level we've not been requested to show", "level", line.Level.String(), "msg", line.Message)
			continue
		}

		// TODO check level is enabled flag
		fmt.Println(style.RenderLine(line, true))
	}
}
