package log

import (
	"encoding/json"
	"fmt"
	"time"

	"github.com/charmbracelet/log"
)

type Level int

const (
	LevelTrace Level = 10
	LevelDebug Level = 20
	LevelInfo  Level = 30
	LevelWarn  Level = 40
	LevelError Level = 50
	LevelFatal Level = 60
)

func (l Level) AsLevel() log.Level {
	// TODO remove

	switch l {
	case LevelTrace:
		return log.DebugLevel // there's no TRACE
	case LevelDebug:
		return log.DebugLevel
	case LevelInfo:
		return log.InfoLevel
	case LevelWarn:
		return log.WarnLevel
	case LevelError:
		return log.ErrorLevel
	case LevelFatal:
		return log.FatalLevel
	default:
		return log.DebugLevel // default to DEBUG
	}
}

func (l Level) String() string {
	switch l {
	case LevelTrace:
		return "TRACE"
	case LevelDebug:
		return "DEBUG"
	case LevelInfo:
		return "INFO"
	case LevelWarn:
		return "WARN"
	case LevelError:
		return "ERROR"
	case LevelFatal:
		return "FATAL"
	default:
		return fmt.Sprintf("Unknown log Level (%d)", l)
	}
}

type Line struct {
	Name       string    `json:"name"`
	Hostname   string    `json:"hostname"`
	Pid        int       `json:"pid"`
	Level      Level     `json:"level"`
	LogContext string    `json:"logContext"`
	Repository string    `json:"repository"`
	Message    string    `json:"msg"`
	Time       time.Time `json:"time"`
	Version    int       `json:"v"`

	Extras map[string]any `json:"-"`
}

func (ll Line) Render() string {
	return ""
}

func (ll *Line) UnmarshalJSON(data []byte) error {
	var tempMap map[string]interface{}
	if err := json.Unmarshal(data, &tempMap); err != nil {
		return err
	}

	// Extract known fields and delete them from the map
	if name, ok := tempMap["name"].(string); ok {
		ll.Name = name
		delete(tempMap, "name")
	}
	if hostname, ok := tempMap["hostname"].(string); ok {
		ll.Hostname = hostname
		delete(tempMap, "hostname")
	}
	if pid, ok := tempMap["pid"].(float64); ok { // JSON numbers are float64
		ll.Pid = int(pid)
		delete(tempMap, "pid")
	}
	if level, ok := tempMap["level"].(float64); ok {
		ll.Level = Level(int(level))
		delete(tempMap, "level")
	}
	if logContext, ok := tempMap["logContext"].(string); ok {
		ll.LogContext = logContext
		delete(tempMap, "logContext")
	}
	if repository, ok := tempMap["repository"].(string); ok {
		ll.Repository = repository
		delete(tempMap, "repository")
	}
	if msg, ok := tempMap["msg"].(string); ok {
		ll.Message = msg
		delete(tempMap, "msg")
	}
	if timeStr, ok := tempMap["time"].(string); ok {
		parsedTime, err := time.Parse(time.RFC3339, timeStr)
		if err != nil {
			return err
		}
		ll.Time = parsedTime
		delete(tempMap, "time")
	}
	if version, ok := tempMap["v"].(float64); ok {
		ll.Version = int(version)
		delete(tempMap, "v")
	}

	ll.Extras = tempMap

	return nil
}
