package style

import (
	"bytes"
	"encoding/json"
	"fmt"
	"regexp"
	"slices"
	"strings"
	"time"

	"github.com/charmbracelet/lipgloss"
	"gitlab.com/tanna.dev/renovate-pretty-log/internal/log"
)

// HACK for https://github.com/charmbracelet/lipgloss/issues/199
var reHackToRemoveInadvertentWhitespaceIntroducedByLipGloss = regexp.MustCompile(`([ \t]*[\n\r]|[ \t]*$|[\n\r]$)`)

// HACK for https://github.com/charmbracelet/lipgloss/issues/199
var hackInadvertentTrailerBytes = []byte{
	27, 91, 48, 109, 10, 27, 91, 51, 56, 59, 50, 59, 49, 56, 54, 59, 49, 54, 54, 59, 49, 50, 55, 109, 27, 91, 48, 109, 10,
}

var LogLineTraceLevelStyle = lipgloss.NewStyle().
	SetString("TRAC").
	Bold(true).
	MaxWidth(4).
	Foreground(lipgloss.Color("#2C78BF"))

var LogLineDebugLevelStyle = lipgloss.NewStyle().
	SetString("DEBU").
	Bold(true).
	MaxWidth(4).
	Foreground(lipgloss.Color("#2C78BF"))

var LogLineInfoLevelStyle = lipgloss.NewStyle().
	SetString("INFO").
	Bold(true).
	MaxWidth(4).
	Foreground(lipgloss.Color("#519F50"))

var LogLineWarnLevelStyle = lipgloss.NewStyle().
	SetString("WARN").
	Bold(true).
	MaxWidth(4).
	Foreground(lipgloss.Color("#FBB829"))

var LogLineErrorLevelStyle = lipgloss.NewStyle().
	SetString("ERRO").
	Bold(true).
	MaxWidth(4).
	Foreground(lipgloss.Color("#EF2F27"))

var LogLineFatalLevelStyle = lipgloss.NewStyle().
	SetString("FATA").
	Bold(true).
	MaxWidth(4).
	Foreground(lipgloss.Color("#F75341"))

var LogLineHasExtrasStyle = lipgloss.NewStyle().
	Foreground(lipgloss.Color("#ff5c8f"))

var LogLineExtraKeyStyle = lipgloss.NewStyle().
	Foreground(lipgloss.Color("#FCE8C3"))

var LogLineExtraValueStyle = lipgloss.NewStyle().
	Foreground(lipgloss.Color("#BAA67F"))

var LogLineTimestampStyle = lipgloss.NewStyle().
	Bold(true).
	MaxWidth(19).
	Foreground(lipgloss.Color("240"))

var LogLineMessageStyle = lipgloss.NewStyle().
	Foreground(lipgloss.Color("#FFFFFF"))

var LogLineExtraIndent = "     "

func LevelStyle(level log.Level) lipgloss.Style {
	switch level {
	case log.LevelTrace:
		return LogLineTraceLevelStyle
	case log.LevelDebug:
		return LogLineDebugLevelStyle
	case log.LevelInfo:
		return LogLineInfoLevelStyle
	case log.LevelWarn:
		return LogLineWarnLevelStyle
	case log.LevelError:
		return LogLineErrorLevelStyle
	case log.LevelFatal:
		return LogLineFatalLevelStyle
	default:
		panic("TODO")
	}
}

func TimestampStyle(timestamp *time.Time) lipgloss.Style {
	return LogLineTimestampStyle.
		SetString(timestamp.Format(time.RFC3339))
}

func MessageStyle(message string) lipgloss.Style {
	return LogLineMessageStyle.
		SetString(message)
}

func RenderLine(line log.Line, withExtras bool) string {
	var parts []string

	out := LevelStyle(line.Level).Render()
	parts = append(parts, out)

	out = TimestampStyle(&line.Time).Render()
	parts = append(parts, out)

	out = MessageStyle(line.Message).Render()
	parts = append(parts, out)

	if withExtras {
		extras := RenderLineExtras(line)
		if extras != "" {
			parts = append(parts, "\n"+extras)
		}
	}

	return strings.Join(parts, " ")
}

func RenderLineExtras(line log.Line) string {
	keys := make([]string, 0, len(line.Extras))
	for k := range line.Extras {
		keys = append(keys, k)
	}
	slices.Sort(keys)

	var buf bytes.Buffer
	jsonEncoder := json.NewEncoder(&buf)
	jsonEncoder.SetIndent(LogLineExtraIndent, "  ")
	/*
	   "go/google.golang.org/protobuf": {
	     "\u003c 1.33.0": "1.33.0"
	   }
	*/
	jsonEncoder.SetEscapeHTML(false)

	var extras []string
	for _, k := range keys {
		v := line.Extras[k]
		out := ""

		switch v := v.(type) {
		case bool,
			float32,
			float64,
			string:
			out = fmt.Sprintf("%#v", v)
		default:
			buf.Reset()
			err := jsonEncoder.Encode(v)
			if err == nil {
				out = buf.String()
			} else {
				out = fmt.Sprintf("%#v", v)
			}
		}

		out = LogLineExtraValueStyle.Render(out)
		out = reHackToRemoveInadvertentWhitespaceIntroducedByLipGloss.ReplaceAllString(out, "\n")
		out = strings.TrimSuffix(out, string(hackInadvertentTrailerBytes))

		extras = append(extras, LogLineExtraIndent+LogLineExtraKeyStyle.Render(k)+": "+out)
	}

	if len(extras) > 0 {
		return strings.Join(extras, "\n")
	}

	return ""
}
